@@ -1,12 +1,16 @@
 """Handlers module
 """
 
-from __future__ import division, absolute_import, print_function
+from __future__ import absolute_import, division, print_function
 
-import sys
 import functools
+import sys
 
-DEFAULT_HANDLERS = []
+from typing import List
+
+import londiste.handler
+
+DEFAULT_HANDLERS: List["londiste.handler.BaseHandler"] = []
 
 
 def handler_args(name, cls):
@@ -15,6 +19,7 @@ def handler_args(name, cls):
     Define successor for handler class cls with func as argument generator
     """
     def wrapper(func):
+        # pylint: disable=unnecessary-dunder-call
         def _init_override(self, table_name, args, dest_table):
             cls.__init__(self, table_name, func(args.copy()), dest_table)
         dct = {'__init__': _init_override, 'handler_name': name}
@@ -22,8 +27,8 @@ def handler_args(name, cls):
         newname = '%s_%s' % (cls.__name__, name.replace('.', '_'))
         newcls = type(newname, (cls,), dct)
         setattr(module, newname, newcls)
-        module.__londiste_handlers__.append(newcls)
-        module.__all__.append(newname)
+        getattr(module, "__londiste_handlers__").append(newcls)
+        getattr(module, "__all__").append(newname)
         return func
     return wrapper
 
@@ -33,3 +38,4 @@ def update(*p):
     in reverse order """
     return functools.reduce(lambda x, y: x.update(y) or x,
                             (p[i] for i in range(len(p) - 1, -1, -1)), {})
+
