@@ -1,8 +1,8 @@
-"""
-Send all events to a DB function.
+"""Send all events to a DB function.
 """
 
 import skytools
+
 from londiste.handler import BaseHandler
 
 __all__ = ['ApplyFuncHandler']
@@ -19,7 +19,8 @@ class ApplyFuncHandler(BaseHandler):
     cur_tick = None
 
     def prepare_batch(self, batch_info, dst_curs):
-        self.cur_tick = batch_info['tick_id']
+        if batch_info is not None:
+            self.cur_tick = batch_info['tick_id']
 
     def process_event(self, ev, sql_queue_func, qfunc_arg):
         """Ignore events for this table"""
@@ -43,4 +44,6 @@ class ApplyFuncHandler(BaseHandler):
 # register handler class
 #------------------------------------------------------------------------------
 
+
 __londiste_handlers__ = [ApplyFuncHandler]
+
