@@ -163,15 +163,17 @@ creating or coping initial data to destination table.  --expect-sync and
 
 import datetime
 import re
-from functools import partial
+import logging
+from typing import Sequence, List, Tuple, Optional
 
 import skytools
-from skytools import quote_ident, quote_fqident, UsageError
-from skytools.dbstruct import TableStruct, T_ALL
+from skytools import UsageError, quote_fqident, quote_ident
+from skytools.dbstruct import T_ALL, TableStruct
+from skytools.basetypes import Cursor
 
 from londiste.handlers import handler_args, update
 from londiste.handlers.shard import ShardHandler
-
+import londiste.util
 
 __all__ = ['Dispatcher']
 
@@ -208,12 +210,12 @@ RETENTION_FUNC = "londiste.drop_obsolete_partitions"
 #------------------------------------------------------------------------------
 
 
-class BaseLoader(object):
-    def __init__(self, table, pkeys, log, conf):
+class BaseLoader:
+    def __init__(self, table: str, pkeys: Sequence[str], log: logging.Logger, conf: skytools.dbdict):
         self.table = table
         self.pkeys = pkeys
         self.log = log
-        self.conf = conf or {}
+        self.conf = conf or skytools.dbdict()
 
     def process(self, op, row):
         raise NotImplementedError()
@@ -224,7 +226,7 @@ class BaseLoader(object):
 
 class DirectLoader(BaseLoader):
     def __init__(self, table, pkeys, log, conf):
-        super(DirectLoader, self).__init__(table, pkeys, log, conf)
+        super().__init__(table, pkeys, log, conf)
         self.data = []
 
     def process(self, op, row):
@@ -255,7 +257,7 @@ class BaseBulkCollectingLoader(BaseLoader):
                 }
 
     def __init__(self, table, pkeys, log, conf):
-        super(BaseBulkCollectingLoader, self).__init__(table, pkeys, log, conf)
+        super().__init__(table, pkeys, log, conf)
         if not self.pkeys:
             raise Exception('non-pk tables not supported: %s' % self.table)
         self.pkey_ev_map = {}
@@ -277,7 +279,7 @@ class BaseBulkCollectingLoader(BaseLoader):
             if len(pk_data) == len(row) and _op == 'U':
                 del self.pkey_ev_map[pk_data]
         except KeyError:
-            raise Exception('unknown event type: %s' % op)
+            raise Exception('unknown event type: %s' % op) from None
 
     def collect_data(self):
         """Collects list of rows into operation hashed dict
@@ -300,8 +302,10 @@ class BaseBulkCollectingLoader(BaseLoader):
 class BaseBulkTempLoader(BaseBulkCollectingLoader):
     """ Provide methods for operating bulk collected events with temp table
     """
+    keys: List[str]
+
     def __init__(self, table, pkeys, log, conf):
-        super(BaseBulkTempLoader, self).__init__(table, pkeys, log, conf)
+        super().__init__(table, pkeys, log, conf)
         # temp table name
         if USE_REAL_TABLE:
             self.temp = self.table + "_loadertmpx"
@@ -315,7 +319,7 @@ class BaseBulkTempLoader(BaseBulkCollectingLoader):
         self.fields = None
         # key fields used in where part, possible to add non pk fields
         # (like dist keys in gp)
-        self.keys = self.pkeys[:]
+        self.keys = list(self.pkeys)
 
     def nonkeys(self):
         """returns fields not in keys"""
@@ -340,7 +344,7 @@ class BaseBulkTempLoader(BaseBulkCollectingLoader):
 
     def insert(self, curs):
         sql = "insert into %s (%s) select %s from %s" % (self.qtable, self._cols(), self._cols(), self.qtemp)
-        return self.logexec(curs, sql)
+        self.logexec(curs, sql)
 
     def update(self, curs):
         qcols = [quote_ident(c) for c in self.nonkeys()]
@@ -354,30 +358,30 @@ class BaseBulkTempLoader(BaseBulkCollectingLoader):
         _set = ", ".join(eqlist)
 
         sql = "update only %s set %s from %s as t where %s" % (self.qtable, _set, self.qtemp, self._where())
-        return self.logexec(curs, sql)
+        self.logexec(curs, sql)
 
     def delete(self, curs):
         sql = "delete from only %s using %s as t where %s" % (self.qtable, self.qtemp, self._where())
-        return self.logexec(curs, sql)
+        self.logexec(curs, sql)
 
     def truncate(self, curs):
-        return self.logexec(curs, "truncate %s" % self.qtemp)
+        self.logexec(curs, "truncate %s" % self.qtemp)
 
     def drop(self, curs):
-        return self.logexec(curs, "drop table %s" % self.qtemp)
+        self.logexec(curs, "drop table %s" % self.qtemp)
 
     def create(self, curs):
         if USE_REAL_TABLE:
             tmpl = "create table %s (like %s)"
         else:
             tmpl = "create temp table %s (like %s) on commit preserve rows"
-        return self.logexec(curs, tmpl % (self.qtemp, self.qtable))
+        self.logexec(curs, tmpl % (self.qtemp, self.qtable))
 
     def analyze(self, curs):
-        return self.logexec(curs, "analyze %s" % self.qtemp)
+        self.logexec(curs, "analyze %s" % self.qtemp)
 
     def process(self, op, row):
-        super(BaseBulkTempLoader, self).process(op, row)
+        super().process(op, row)
         # TODO: maybe one assignment is enough?
         self.fields = row.keys()
 
@@ -386,7 +390,7 @@ class BulkLoader(BaseBulkTempLoader):
     """ Collects events to and loads bulk data using copy and temp tables
     """
     def __init__(self, table, pkeys, log, conf):
-        super(BulkLoader, self).__init__(table, pkeys, log, conf)
+        super().__init__(table, pkeys, log, conf)
         self.method = self.conf['method']
         self.run_analyze = self.conf['analyze']
         self.dist_fields = None
@@ -396,7 +400,7 @@ class BulkLoader(BaseBulkTempLoader):
     def process(self, op, row):
         if self.method == METH_INSERT and op != 'I':
             raise Exception('%s not supported by method insert' % op)
-        super(BulkLoader, self).process(op, row)
+        super().process(op, row)
 
     def process_delete(self, curs, op_map):
         """Process delete list"""
@@ -559,7 +563,7 @@ LOADERS = {'direct': DirectLoader, 'bulk': BulkLoader}
 # ROW HANDLERS
 #------------------------------------------------------------------------------
 
-class RowHandler(object):
+class RowHandler:
     def __init__(self, log):
         self.log = log
         self.table_map = {}
@@ -571,7 +575,7 @@ class RowHandler(object):
         try:
             self.table_map[table].process(op, row)
         except KeyError:
-            raise Exception("No loader for table %s" % table)
+            raise Exception("No loader for table %s" % table) from None
 
     def flush(self, curs):
         for ldr in self.table_map.values():
@@ -589,7 +593,7 @@ class KeepAllRowHandler(RowHandler):
             op = 'I'
         elif op == 'D':
             return
-        super(KeepAllRowHandler, self).process(table, op, row)
+        super().process(table, op, row)
 
 
 class KeepLatestRowHandler(RowHandler):
@@ -600,12 +604,12 @@ class KeepLatestRowHandler(RowHandler):
         Makes sense only for partitioned tables.
         """
         if op == 'U':
-            super(KeepLatestRowHandler, self).process(table, 'D', row)
-            super(KeepLatestRowHandler, self).process(table, 'I', row)
+            super().process(table, 'D', row)
+            super().process(table, 'I', row)
         elif op == 'I':
-            super(KeepLatestRowHandler, self).process(table, 'I', row)
+            super().process(table, 'I', row)
         elif op == 'D':
-            super(KeepLatestRowHandler, self).process(table, 'D', row)
+            super().process(table, 'D', row)
 
 
 ROW_HANDLERS = {'plain': RowHandler,
@@ -633,7 +637,7 @@ class Dispatcher(ShardHandler):
         # compat for dest-table
         dest_table = args.get('table', dest_table)
 
-        super(Dispatcher, self).__init__(table_name, args, dest_table)
+        super().__init__(table_name, args, dest_table)
 
         # show args
         self.log.debug("dispatch.init: table_name=%r, args=%r", table_name, args)
@@ -670,7 +674,7 @@ class Dispatcher(ShardHandler):
 
     def get_config(self):
         """Processes args dict"""
-        conf = super(Dispatcher, self).get_config()
+        conf = super().get_config()
         # set table mode
         conf.table_mode = self.get_arg('table_mode', TABLE_MODES)
         conf.analyze = self.get_arg('analyze', [0, 1])
@@ -721,17 +725,12 @@ class Dispatcher(ShardHandler):
     def _validate_hash_key(self):
         pass  # no need for hash key when not sharding
 
-    def reset(self):
-        """Called before starting to process a batch.
-        Should clean any pending data."""
-        super(Dispatcher, self).reset()
-
     def prepare_batch(self, batch_info, dst_curs):
         """Called on first event for this table in current batch."""
-        if self.conf.table_mode != 'ignore':
+        if batch_info is not None and self.conf.table_mode != 'ignore':
             self.batch_info = batch_info
             self.dst_curs = dst_curs
-        super(Dispatcher, self).prepare_batch(batch_info, dst_curs)
+        super().prepare_batch(batch_info, dst_curs)
 
     def filter_data(self, data):
         """Process with fields skip and map"""
@@ -800,7 +799,7 @@ class Dispatcher(ShardHandler):
         """Called when batch finishes."""
         if self.conf.table_mode != 'ignore':
             self.row_handler.flush(dst_curs)
-        #super(Dispatcher, self).finish_batch(batch_info, dst_curs)
+        #super().finish_batch(batch_info, dst_curs)
 
     def get_part_name(self):
         # if custom part name template given, use it
@@ -938,11 +937,6 @@ class Dispatcher(ShardHandler):
             self.log.info("Ignored table: %s", partition_table)
         return res
 
-    def get_copy_condition(self, src_curs, dst_curs):
-        """ Prepare where condition for copy and replay filtering.
-        """
-        return super(Dispatcher, self).get_copy_condition(src_curs, dst_curs)
-
     def real_copy(self, tablename, src_curs, dst_curs, column_list):
         """do actual table copy and return tuple with number of bytes and rows
         copied
@@ -971,6 +965,49 @@ class Dispatcher(ShardHandler):
                                   dst_column_list=_dst_cols,
                                   write_hook=_write_hook)
 
+    def real_copy_threaded(
+        self,
+        src_real_table: str,
+        src_curs: Cursor,
+        dst_db_connstr: str,
+        column_list: Sequence[str],
+        config_file: str,
+        config_section: str,
+        parallel: int = 1,
+    ) -> Tuple[int, int]:
+        with skytools.connect_database(dst_db_connstr) as dst_db:
+            with dst_db.cursor() as dst_curs:
+                condition = self.get_copy_condition(src_curs, dst_curs)
+            dst_db.commit()
+
+        _src_cols = _dst_cols = column_list
+        if self.conf.skip_fields:
+            _src_cols = [col for col in column_list
+                         if col not in self.conf.skip_fields]
+            _dst_cols = _src_cols
+
+        if self.conf.field_map:
+            _src_cols = [col for col in _src_cols if col in self.conf.field_map]
+            _dst_cols = [self.conf.field_map[col] for col in _src_cols]
+
+        _write_hook: Optional[londiste.util.WriteHook]
+        if self.encoding_validator:
+            def _write_hook(obj, data):
+                return self.encoding_validator.validate_copy(data, _src_cols, src_real_table)
+        else:
+            _write_hook = None
+
+        return londiste.util.full_copy_parallel(
+            src_real_table, src_curs,
+            dst_db_connstr=dst_db_connstr,
+            dst_tablename=self.dest_table,
+            condition=condition,
+            column_list=_src_cols,
+            dst_column_list=_dst_cols,
+            write_hook=_write_hook,
+            parallel=parallel,
+        )
+
 
 # add arguments' description to handler's docstring
 def _install_handler_docstrings(dst_cls):
@@ -981,6 +1018,7 @@ def _install_handler_docstrings(dst_cls):
         if found:
             dst_cls._doc_ += "\n" + line
 
+
 _install_handler_docstrings(Dispatcher)
 
 #------------------------------------------------------------------------------
@@ -991,12 +1029,6 @@ _install_handler_docstrings(Dispatcher)
 __londiste_handlers__ = [Dispatcher]
 
 
-#------------------------------------------------------------------------------
-# helper function for creating dispatchers with different default values
-#------------------------------------------------------------------------------
-
-handler_args = partial(handler_args, cls=Dispatcher)
-
 #------------------------------------------------------------------------------
 # build set of handlers with different default values for easier use
 #------------------------------------------------------------------------------
@@ -1042,7 +1074,7 @@ def _generate_handlers():
                 def create_handler(_handler_name, _load_dict, _period_dict, _mode_dict):
                     default = update(_mode_dict, _period_dict, _load_dict, BASE)
 
-                    @handler_args(_handler_name)
+                    @handler_args(_handler_name, Dispatcher)
                     def handler_func(args):
                         return update(args, default)
 
@@ -1053,18 +1085,22 @@ def _generate_handlers():
                 defs = update(mode_dict, period_dict, load_dict, BASE)
                 set_handler_doc(hcls, defs)
 
+
 _generate_handlers()
 
 
-@handler_args('bulk_direct')
+@handler_args('bulk_direct', Dispatcher)
 def bulk_direct_handler(args):
     return update(args, {'load_mode': 'bulk', 'table_mode': 'direct'})
 
+
 set_handler_doc(__londiste_handlers__[-1], {'load_mode': 'bulk', 'table_mode': 'direct'})
 
 
-@handler_args('direct')
+@handler_args('direct', Dispatcher)
 def direct_handler(args):
     return update(args, {'load_mode': 'direct', 'table_mode': 'direct'})
 
+
 set_handler_doc(__londiste_handlers__[-1], {'load_mode': 'direct', 'table_mode': 'direct'})
+
