@@ -1,20 +1,27 @@
-"""Basic replication core."""
+"""Basic replication core.
+"""
 
-from __future__ import division, absolute_import, print_function
-
-import sys
 import os
+import sys
 import time
+import fnmatch
+
+from logging import Logger
+from typing import List, Optional, Dict, Sequence, Mapping
+
 import skytools
 
+from skytools.basetypes import DictRow
+
+from pgq.event import Event
 from pgq.cascade.worker import CascadedWorker
 
-from londiste.handler import load_handler_modules, build_handler
-from londiste.exec_attrs import ExecAttrs
+from .exec_attrs import ExecAttrs
+from .handler import build_handler, load_handler_modules, BaseHandler
 
 __all__ = ['Replicator', 'TableState',
-    'TABLE_MISSING', 'TABLE_IN_COPY', 'TABLE_CATCHING_UP',
-    'TABLE_WANNA_SYNC', 'TABLE_DO_SYNC', 'TABLE_OK']
+           'TABLE_MISSING', 'TABLE_IN_COPY', 'TABLE_CATCHING_UP',
+           'TABLE_WANNA_SYNC', 'TABLE_DO_SYNC', 'TABLE_OK']
 
 # state                 # owner - who is allowed to change
 TABLE_MISSING = 0       # main
@@ -28,9 +35,20 @@ SYNC_OK = 0    # continue with batch
 SYNC_LOOP = 1  # sleep, try again
 SYNC_EXIT = 2  # nothing to do, exit script
 
-MAX_PARALLEL_COPY = 8 # default number of allowed max parallel copy processes
+MAX_PARALLEL_COPY = 8  # default number of allowed max parallel copy processes
 
-class Counter(object):
+
+def is_data_event(ev: Event) -> bool:
+    """Is it insert/update/delete for one table?
+    """
+    if ev.type in ('I', 'U', 'D'):
+        return True
+    elif ev.type[:2] in ('I:', 'U:', 'D:', '{"'):
+        return True
+    return False
+
+
+class Counter:
     """Counts table statuses."""
 
     missing = 0
@@ -40,13 +58,17 @@ class Counter(object):
     do_sync = 0
     ok = 0
 
-    def __init__(self, tables):
+    def __init__(self, tables: List["TableState"], copy_method_map: Dict[str, Optional[int]]) -> None:
         """Counts and sanity checks."""
         for t in tables:
             if t.state == TABLE_MISSING:
                 self.missing += 1
             elif t.state == TABLE_IN_COPY:
-                self.copy += 1
+                nthreads = copy_method_map[t.name]
+                if nthreads is None:
+                    self.copy += 1
+                else:
+                    self.copy += nthreads
             elif t.state == TABLE_CATCHING_UP:
                 self.catching_up += 1
             elif t.state == TABLE_WANNA_SYNC:
@@ -56,12 +78,32 @@ class Counter(object):
             elif t.state == TABLE_OK:
                 self.ok += 1
 
-    def get_copy_count(self):
+    def get_copy_count(self) -> int:
         return self.copy + self.catching_up + self.wanna_sync + self.do_sync
 
-class TableState(object):
+
+class TableState:
     """Keeps state about one table."""
-    def __init__(self, name, log):
+
+    name: str
+    dest_table: str
+    log: Logger
+    state: int
+    last_snapshot_tick: Optional[int]
+    str_snapshot: Optional[str]
+    from_snapshot: Optional[skytools.Snapshot]
+    sync_tick_id: Optional[int]
+    ok_batch_count: int
+    last_tick: Optional[int]
+    table_attrs: Mapping[str, Optional[str]]
+    copy_role: Optional[str]
+    dropped_ddl: Optional[str]
+    plugin: Optional[BaseHandler]
+    changed: int
+    copy_pos: int
+    max_parallel_copy: int
+
+    def __init__(self, name: str, log: Logger) -> None:
         """Init TableState for one table."""
         self.name = name
         self.dest_table = name
@@ -85,7 +127,7 @@ class TableState(object):
         # max number of parallel copy processes allowed
         self.max_parallel_copy = MAX_PARALLEL_COPY
 
-    def forget(self):
+    def forget(self) -> None:
         """Reset all info."""
         self.state = TABLE_MISSING
         self.last_snapshot_tick = None
@@ -100,7 +142,7 @@ class TableState(object):
         self.copy_pos = 0
         self.max_parallel_copy = MAX_PARALLEL_COPY
 
-    def change_snapshot(self, str_snapshot, tag_changed=1):
+    def change_snapshot(self, str_snapshot: Optional[str], tag_changed: int = 1) -> None:
         """Set snapshot."""
         if self.str_snapshot == str_snapshot:
             return
@@ -116,7 +158,7 @@ class TableState(object):
             self.last_tick = None
             self.changed = 1
 
-    def change_state(self, state, tick_id=None):
+    def change_state(self, state: int, tick_id: Optional[int] = None) -> None:
         """Set state."""
         if self.state == state and self.sync_tick_id == tick_id:
             return
@@ -125,7 +167,7 @@ class TableState(object):
         self.changed = 1
         self.log.debug("%s: change_state to %s", self.name, self.render_state())
 
-    def render_state(self):
+    def render_state(self) -> Optional[str]:
         """Make a string to be stored in db."""
 
         if self.state == TABLE_MISSING:
@@ -135,13 +177,14 @@ class TableState(object):
         elif self.state == TABLE_CATCHING_UP:
             return 'catching-up'
         elif self.state == TABLE_WANNA_SYNC:
-            return 'wanna-sync:%d' % self.sync_tick_id
+            return 'wanna-sync:%d' % (self.sync_tick_id or 0)
         elif self.state == TABLE_DO_SYNC:
-            return 'do-sync:%d' % self.sync_tick_id
+            return 'do-sync:%d' % (self.sync_tick_id or 0)
         elif self.state == TABLE_OK:
             return 'ok'
+        return None
 
-    def parse_state(self, merge_state):
+    def parse_state(self, merge_state) -> int:
         """Read state from string."""
 
         state = -1
@@ -169,7 +212,7 @@ class TableState(object):
 
         return state
 
-    def loaded_state(self, row):
+    def loaded_state(self, row: DictRow) -> None:
         """Update object with info from db."""
 
         self.log.debug("loaded_state: %s: %s / %s",
@@ -187,23 +230,25 @@ class TableState(object):
             self.changed = 1
 
         self.copy_pos = int(row.get('copy_pos', '0'))
-        self.max_parallel_copy = int(self.table_attrs.get('max_parallel_copy',
-                                                        self.max_parallel_copy))
+
+        max_parallel_copy = self.table_attrs.get('max_parallel_copy')
+        if max_parallel_copy:
+            self.max_parallel_copy = int(max_parallel_copy)
 
         if 'dest_table' in row and row['dest_table']:
             self.dest_table = row['dest_table']
         else:
             self.dest_table = self.name
 
-        hstr = self.table_attrs.get('handlers', '') # compat
+        hstr = self.table_attrs.get('handlers', '')  # compat
         hstr = self.table_attrs.get('handler', hstr)
         self.plugin = build_handler(self.name, hstr, self.dest_table)
 
-    def max_parallel_copies_reached(self):
-        return self.max_parallel_copy and\
-                    self.copy_pos >= self.max_parallel_copy
+    def max_parallel_copies_reached(self) -> bool:
+        return self.max_parallel_copy is not None and \
+            self.copy_pos >= self.max_parallel_copy
 
-    def interesting(self, ev, tick_id, copy_thread, copy_table_name):
+    def interesting(self, ev: Event, tick_id: int, copy_thread: bool, copy_table_name: str) -> bool:
         """Check if table wants this event."""
 
         if copy_thread:
@@ -234,7 +279,7 @@ class TableState(object):
                 self.change_snapshot(None)
         return True
 
-    def gc_snapshot(self, copy_thread, prev_tick, cur_tick, no_lag):
+    def gc_snapshot(self, copy_thread: bool, prev_tick: int, cur_tick: int, no_lag: bool) -> None:
         """Remove attached snapshot if possible.
 
         If the event processing is in current moment, the snapshot
@@ -255,7 +300,7 @@ class TableState(object):
                 return
         else:
             if self.state != TABLE_OK:
-                return False
+                return
 
         # aquire last tick
         if not self.last_snapshot_tick:
@@ -267,9 +312,12 @@ class TableState(object):
         if self.last_snapshot_tick < prev_tick:
             self.change_snapshot(None)
 
-    def get_plugin(self):
+    def get_plugin(self) -> BaseHandler:
+        if not self.plugin:
+            raise ValueError("no handler set")
         return self.plugin
 
+
 class Replicator(CascadedWorker):
     """Replication core.
 
@@ -287,8 +335,15 @@ class Replicator(CascadedWorker):
         # how many tables can be copied in parallel
         #parallel_copies = 1
 
+        # glob patterns for table names: archive.*, public.*
+        #threaded_copy_tables =
+        # number of threads in pool
+        #threaded_copy_pool_size = 1
+
         # accept only events for locally present tables
-        #local_only = true
+        #local_only = false
+        # do not load EXECUTE events from source queue when local_only is active
+        #local_only_drop_execute = false
 
         ## compare/repair
         # max amount of time table can be locked
@@ -315,23 +370,49 @@ class Replicator(CascadedWorker):
 
         # connect string for existing node to use as provider
         #initial_provider_location =
+
+        # filter for table/seq registration
+        #register_only_tables =
+        #register_only_seqs =
+        #register_skip_tables = s.a, s.b, s.c
+        #register_skip_seqs =
     """
 
     # batch info
-    cur_tick = 0
-    prev_tick = 0
-    copy_table_name = None # filled by Copytable()
-    sql_list = []
+    cur_tick: int = 0
+    prev_tick: int = 0
+    copy_table_name: Optional[str] = None  # filled by Copytable()
+    sql_list: List[str] = []
 
-    current_event = None
+    current_event: Optional[Event] = None
 
-    def __init__(self, args):
+    threaded_copy_tables: Sequence[str]
+    threaded_copy_pool_size: int
+    copy_method_map: Dict[str, Optional[int]]
+
+    register_only_tables: Optional[Sequence[str]] = None
+    register_only_seqs: Optional[Sequence[str]] = None
+    register_skip_tables: Optional[Sequence[str]] = None
+    register_skip_seqs: Optional[Sequence[str]] = None
+
+    local_only: bool = False
+    local_only_drop_execute: bool = False
+
+    table_list: List[TableState]
+    table_map: Dict[str, TableState]
+    used_plugins: Dict[str, BaseHandler]
+
+    def __init__(self, args: Sequence[str]):
         """Replication init."""
-        super(Replicator, self).__init__('londiste', 'db', args)
+        super().__init__('londiste', 'db', args)
 
         self.table_list = []
         self.table_map = {}
 
+        self.threaded_copy_tables = self.cf.getlist('threaded_copy_tables', [])
+        self.threaded_copy_pool_size = self.cf.getint('threaded_copy_pool_size', 1)
+        self.copy_method_map = {}
+
         self.copy_thread = 0
         self.set_name = self.queue_name
         self.used_plugins = {}
@@ -342,11 +423,41 @@ class Replicator(CascadedWorker):
 
         self.consumer_filter = None
 
+        self.register_only_tables = self.cf.getlist("register_only_tables", [])
+        self.register_only_seqs = self.cf.getlist("register_only_seqs", [])
+        self.register_skip_tables = self.cf.getlist("register_skip_tables", [])
+        self.register_skip_seqs = self.cf.getlist("register_skip_seqs", [])
+
+        self.local_only = self.cf.getboolean('local_only', False)
+        self.local_only_drop_execute = self.cf.getboolean('local_only_drop_execute', False)
+
     def reload(self):
-        super(Replicator, self).reload()
+        super().reload()
 
         load_handler_modules(self.cf)
 
+        self.threaded_copy_tables = self.cf.getlist('threaded_copy_tables', [])
+        self.threaded_copy_pool_size = self.cf.getint('threaded_copy_pool_size', 1)
+        self.copy_method_map = {}
+
+        self.register_only_tables = self.cf.getlist("register_only_tables", [])
+        self.register_only_seqs = self.cf.getlist("register_only_seqs", [])
+        self.register_skip_tables = self.cf.getlist("register_skip_tables", [])
+        self.register_skip_seqs = self.cf.getlist("register_skip_seqs", [])
+
+        self.local_only = self.cf.getboolean('local_only', False)
+        self.local_only_drop_execute = self.cf.getboolean('local_only_drop_execute', False)
+
+    def fill_copy_method(self):
+        for table_name in self.table_map:
+            if table_name not in self.copy_method_map:
+                for pat in self.threaded_copy_tables:
+                    if fnmatch.fnmatchcase(table_name, pat):
+                        self.copy_method_map[table_name] = self.threaded_copy_pool_size
+                        break
+                if table_name not in self.copy_method_map:
+                    self.copy_method_map[table_name] = None
+
     def connection_hook(self, dbname, db):
         if dbname == 'db':
             curs = db.cursor()
@@ -397,7 +508,7 @@ class Replicator(CascadedWorker):
         # the cascade-consumer can save last tick and commit.
 
         self.sql_list = []
-        super(Replicator, self).process_remote_batch(src_db, tick_id, ev_list, dst_db)
+        super().process_remote_batch(src_db, tick_id, ev_list, dst_db)
         self.flush_sql(dst_curs)
 
         for p in self.used_plugins.values():
@@ -407,24 +518,6 @@ class Replicator(CascadedWorker):
         # finalize table changes
         self.save_table_state(dst_curs)
 
-        # store event filter
-        if self.cf.getboolean('local_only', False):
-            # create list of tables
-            if self.copy_thread:
-                _filterlist = skytools.quote_literal(self.copy_table_name)
-            else:
-                _filterlist = ','.join(map(skytools.quote_literal, self.table_map.keys()))
-
-            # build filter
-            meta = "(ev_type like 'pgq.%' or ev_type like 'londiste.%')"
-            if _filterlist:
-                self.consumer_filter = "(%s or (ev_extra1 in (%s)))" % (meta, _filterlist)
-            else:
-                self.consumer_filter = meta
-        else:
-            # no filter
-            self.consumer_filter = None
-
     def sync_tables(self, src_db, dst_db):
         """Table sync loop.
 
@@ -432,8 +525,8 @@ class Replicator(CascadedWorker):
         return one of SYNC_* constants."""
 
         self.log.debug('Sync tables')
-        while 1:
-            cnt = Counter(self.table_list)
+        while True:
+            cnt = Counter(self.table_list, self.copy_method_map)
             if self.copy_thread:
                 res = self.sync_from_copy_thread(cnt, src_db, dst_db)
             else:
@@ -513,7 +606,7 @@ class Replicator(CascadedWorker):
                         self.log.warning("Table %s not available on provider", t.name)
                         continue
                     pt = pmap[t.name]
-                    if pt.state != TABLE_OK: # or pt.custom_snapshot: # FIXME: does snapsnot matter?
+                    if pt.state != TABLE_OK:  # or pt.custom_snapshot: # FIXME: does snapsnot matter?
                         self.log.info("Table %s not OK on provider, waiting", t.name)
                         continue
 
@@ -562,7 +655,7 @@ class Replicator(CascadedWorker):
                 return SYNC_OK
             else:
                 self.log.error("copy_sync: cur_tick=%d sync_tick=%d",
-                                self.cur_tick, t.sync_tick_id)
+                               self.cur_tick, t.sync_tick_id)
                 raise Exception('Invalid table state')
         elif t.state == TABLE_WANNA_SYNC:
             # wait for main thread to react
@@ -613,7 +706,6 @@ class Replicator(CascadedWorker):
         dst_curs.execute("analyze " + skytools.quote_fqident(ts.name))
         dst_db.commit()
 
-
     def do_copy(self, tbl, src_db, dst_db):
         """Callback for actual copy implementation."""
         raise Exception('do_copy not implemented')
@@ -621,15 +713,16 @@ class Replicator(CascadedWorker):
     def process_remote_event(self, src_curs, dst_curs, ev):
         """handle one event"""
 
-        self.log.debug("New event: id=%s / type=%s / data=%s / extra1=%s", ev.id, ev.type, ev.data, ev.extra1)
+        self.log.debug(
+            "New event: id=%s / type=%s / data=%s / extra1=%s / extra2=%r / extra3=%r",
+            ev.id, ev.type, ev.data, ev.extra1, ev.extra2, ev.extra3
+        )
 
         # set current_event only if processing them one-by-one
         if self.work_state < 0:
             self.current_event = ev
 
-        if ev.type in ('I', 'U', 'D'):
-            self.handle_data_event(ev, dst_curs)
-        elif ev.type[:2] in ('I:', 'U:', 'D:', '{"'):
+        if is_data_event(ev):
             self.handle_data_event(ev, dst_curs)
         elif ev.type == "R":
             self.flush_sql(dst_curs)
@@ -650,7 +743,7 @@ class Replicator(CascadedWorker):
             self.flush_sql(dst_curs)
             self.update_seq(dst_curs, ev)
         else:
-            super(Replicator, self).process_remote_event(src_curs, dst_curs, ev)
+            super().process_remote_event(src_curs, dst_curs, ev)
 
         # no point keeping it around longer
         self.current_event = None
@@ -685,6 +778,7 @@ class Replicator(CascadedWorker):
         except KeyError:
             p = t.get_plugin()
             self.used_plugins[ev.extra1] = p
+            p.prepare_batch(self.batch_info, dst_curs)
 
         if p.conf.get('ignore_truncate'):
             self.log.info("ignoring truncate for %s", fqname)
@@ -770,6 +864,10 @@ class Replicator(CascadedWorker):
     def add_set_table(self, dst_curs, tbl):
         """There was new table added to root, remember it."""
 
+        if self.register_only_tables and tbl not in self.register_only_tables:
+            return
+        if self.register_skip_tables and tbl in self.register_skip_tables:
+            return
         q = "select londiste.global_add_table(%s, %s)"
         dst_curs.execute(q, [self.set_name, tbl])
 
@@ -788,6 +886,30 @@ class Replicator(CascadedWorker):
         q = "select londiste.global_remove_seq(%s, %s)"
         dst_curs.execute(q, [self.set_name, seq])
 
+    def setup_local_only_filter(self):
+        # store event filter
+        if self.local_only:
+            # create list of tables
+            if self.copy_thread:
+                _filterlist = skytools.quote_literal(self.copy_table_name)
+            else:
+                _filterlist = ','.join(map(skytools.quote_literal, self.table_map.keys()))
+
+            # build filter
+            cond_list = [
+                "ev_type like 'pgq.%'",
+                "ev_type like 'londiste.%'",
+            ]
+            if not self.local_only_drop_execute:
+                cond_list.append("ev_type = 'EXECUTE'")
+            if _filterlist:
+                cond_list.append(f"ev_extra1 in ({_filterlist})")
+            expr = " or ".join(cond_list)
+            self.consumer_filter = f"({expr})"
+        else:
+            # no filter
+            self.consumer_filter = None
+
     def load_table_state(self, curs):
         """Load table state from database.
 
@@ -813,6 +935,19 @@ class Replicator(CascadedWorker):
         self.table_list = new_list
         self.table_map = new_map
 
+        self.fill_copy_method()
+        self.setup_local_only_filter()
+
+    def refresh_state(self, dst_db, full_logic=True):
+        res = super().refresh_state(dst_db, full_logic=full_logic)
+
+        # make sure local_only filter is loaded on boot
+        if self.local_only and self.consumer_filter is None:
+            self.load_table_state(dst_db.cursor())
+            dst_db.commit()
+
+        return res
+
     def get_state_map(self, curs):
         """Get dict of table states."""
 
@@ -844,7 +979,7 @@ class Replicator(CascadedWorker):
                 continue
             merge_state = t.render_state()
             self.log.info("storing state of %s: copy:%d new_state:%s",
-                            t.name, self.copy_thread, merge_state)
+                          t.name, self.copy_thread, merge_state)
             q = "select londiste.local_set_table_state(%s, %s, %s, %s)"
             curs.execute(q, [self.set_name,
                              t.name, t.str_snapshot, merge_state])
@@ -936,15 +1071,40 @@ class Replicator(CascadedWorker):
     def restore_fkeys(self, dst_db):
         """Restore fkeys that have both tables on sync."""
         dst_curs = dst_db.cursor()
+
+        # NOT VALID appreared in 9.1
+        q = "select londiste.version() as ext_version, current_setting('server_version_num')::int < 90100 as compat"
+        dst_curs.execute(q)
+        info = dst_curs.fetchone()
+        if info[0]:
+            ext_version = [int(v) for v in info[0].split('.')]
+            do_compat_restore = ext_version < [3, 7] or info[1]
+        else:
+            do_compat_restore = True
+
         # restore fkeys -- one at a time
         q = "select * from londiste.get_valid_pending_fkeys(%s)"
         dst_curs.execute(q, [self.set_name])
         fkey_list = dst_curs.fetchall()
+        dst_db.commit()
+
         for row in fkey_list:
-            self.log.info('Creating fkey: %(fkey_name)s (%(from_table)s --> %(to_table)s)', row)
-            q2 = "select londiste.restore_table_fkey(%(from_table)s, %(fkey_name)s)"
-            dst_curs.execute(q2, row)
-            dst_db.commit()
+            self.log.info('Creating fkey: %s (%s --> %s)', row['fkey_name'], row['from_table'], row['to_table'])
+            if do_compat_restore:
+                q2 = "select londiste.restore_table_fkey(%(from_table)s, %(fkey_name)s)"
+                dst_curs.execute(q2, row)
+                dst_db.commit()
+            else:
+                q3 = "select londiste.restore_table_fkey(%(from_table)s, %(fkey_name)s, true)"
+                done = False
+                while not done:
+                    dst_curs.execute(q3, row)
+                    sql = dst_curs.fetchone()[0]
+                    if sql:
+                        dst_curs.execute(sql)
+                    else:
+                        done = True
+                    dst_db.commit()
 
     def drop_fkeys(self, dst_db, table_name):
         """Drop all foreign keys to and from this table.
@@ -965,7 +1125,7 @@ class Replicator(CascadedWorker):
     def process_root_node(self, dst_db):
         """On root node send seq changes to queue."""
 
-        super(Replicator, self).process_root_node(dst_db)
+        super().process_root_node(dst_db)
 
         q = "select * from londiste.root_check_seqs(%s)"
         self.exec_cmd(dst_db, q, [self.queue_name])
@@ -980,11 +1140,28 @@ class Replicator(CascadedWorker):
         self.exec_cmd(dst_curs, q, [self.queue_name, seq, val])
 
     def copy_event(self, dst_curs, ev, filtered_copy):
+        # filtered_copy means merge-leaf
         # send only data events down (skipping seqs also)
         if filtered_copy:
             if ev.type[:9] in ('londiste.',):
                 return
-        super(Replicator, self).copy_event(dst_curs, ev, filtered_copy)
+
+        if is_data_event(ev):
+            t = self.get_table_by_name(ev.extra1)
+            if t:
+                try:
+                    p = self.used_plugins[ev.extra1]
+                except KeyError:
+                    p = t.get_plugin()
+                    self.used_plugins[ev.extra1] = p
+                    p.prepare_batch(None, dst_curs)
+
+                # handler may rewrite or drop the event
+                ev = p.get_copy_event(ev, self.queue_name)
+                if ev is None:
+                    return
+
+        super().copy_event(dst_curs, ev, filtered_copy)
 
     def exception_hook(self, det, emsg):
         # add event info to error message
@@ -992,8 +1169,10 @@ class Replicator(CascadedWorker):
             ev = self.current_event
             info = "[ev_id=%d,ev_txid=%d] " % (ev.ev_id, ev.ev_txid)
             emsg = info + emsg
-        super(Replicator, self).exception_hook(det, emsg)
+        super().exception_hook(det, emsg)
+
 
 if __name__ == '__main__':
     script = Replicator(sys.argv[1:])
     script.start()
+
