@@ -3,18 +3,20 @@
 For internal usage.
 """
 
-from __future__ import division, absolute_import, print_function
-
 import sys
 import time
-import skytools
 
-from skytools.dbstruct import TableStruct, T_CONSTRAINT, T_INDEX, T_RULE, T_PARENT
+import skytools
+from skytools.dbstruct import (
+    T_CONSTRAINT, T_INDEX, T_PARENT, T_RULE, TableStruct,
+)
+
+from londiste.playback import TABLE_CATCHING_UP, TABLE_OK, Replicator
 from londiste.util import find_copy_source
-from londiste.playback import Replicator, TABLE_OK, TABLE_CATCHING_UP
 
 __all__ = ['CopyTable']
 
+
 class CopyTable(Replicator):
     """Table copy thread implementation."""
 
@@ -26,7 +28,7 @@ class CopyTable(Replicator):
         in same process.
         """
 
-        super(CopyTable, self).__init__(args)
+        super().__init__(args)
 
         if not copy_thread:
             raise Exception("Combined copy not supported")
@@ -62,19 +64,19 @@ class CopyTable(Replicator):
         src_curs = src_db.cursor()
         dst_curs = dst_db.cursor()
 
-        while 1:
+        while True:
             if tbl_stat.copy_role == 'wait-copy':
                 self.log.info('waiting for first partition to initialize copy')
             elif tbl_stat.max_parallel_copies_reached():
                 self.log.info('number of max parallel copies (%s) reached',
-                                tbl_stat.max_parallel_copy)
+                              tbl_stat.max_parallel_copy)
             else:
                 break
             time.sleep(10)
             tbl_stat = self.reload_table_stat(dst_curs, tbl_stat.name)
             dst_db.commit()
 
-        while 1:
+        while True:
             pmap = self.get_state_map(src_db.cursor())
             src_db.commit()
             if tbl_stat.name not in pmap:
@@ -137,7 +139,7 @@ class CopyTable(Replicator):
 
         # drop unnecessary stuff
         if cmode > 0:
-            objs = T_CONSTRAINT | T_INDEX | T_RULE | T_PARENT # | T_TRIGGER
+            objs = T_CONSTRAINT | T_INDEX | T_RULE | T_PARENT  # | T_TRIGGER
             dst_struct.drop(dst_curs, objs, log=self.log)
 
             # drop data
@@ -161,10 +163,26 @@ class CopyTable(Replicator):
                 dst_db.commit()
                 tbl_stat.dropped_ddl = ddl
 
-        # do truncate & copy
-        self.log.info("%s: start copy", tbl_stat.name)
+        use_threads = False
+        threaded_copy_pool_size = self.copy_method_map.get(tbl_stat.name)
+        if threaded_copy_pool_size is not None:
+            use_threads = True
+
+        # do copy
         p = tbl_stat.get_plugin()
-        stats = p.real_copy(src_real_table, src_curs, dst_curs, common_cols)
+        if use_threads:
+            self.log.info("%s: start threaded copy", tbl_stat.name)
+            dst_db_connstr = self.db_cache["db"].loc
+            dst_db.commit()
+            stats = p.real_copy_threaded(
+                src_real_table, src_curs, dst_db_connstr, common_cols,
+                config_file=self.cf.filename,
+                config_section=self.cf.main_section,
+                parallel=threaded_copy_pool_size,
+            )
+        else:
+            self.log.info("%s: start copy", tbl_stat.name)
+            stats = p.real_copy(src_real_table, src_curs, dst_curs, common_cols)
         if stats:
             self.log.info("%s: copy finished: %d bytes, %d rows",
                           tbl_stat.name, stats[0], stats[1])
@@ -228,7 +246,7 @@ class CopyTable(Replicator):
             # check if needed? (table, not existing reg)
             self.register_copy_consumer()
             self.reg_ok = True
-        return super(CopyTable, self).work()
+        return super().work()
 
     def register_copy_consumer(self):
         dst_db = self.get_database('db')
@@ -254,8 +272,10 @@ class CopyTable(Replicator):
         # do we have node here?
         if 'copy_node' in attrs:
             if attrs['copy_node'] == '?':
-                source_node, source_location, ___wname = find_copy_source(self,
-                        self.queue_name, self.copy_table_name, source_node, source_location)
+                source_node, source_location, ___wname = find_copy_source(
+                    self, self.queue_name, self.copy_table_name,
+                    source_node, source_location,
+                )
             else:
                 # take node from attrs
                 source_node = attrs['copy_node']
@@ -268,6 +288,8 @@ class CopyTable(Replicator):
         self.log.info("Using '%s' as source node", source_node)
         self.register_consumer(source_location)
 
+
 if __name__ == '__main__':
     script = CopyTable(sys.argv[1:])
     script.start()
+
